-module(kif).
-compile(export_all).
-author('kapolnai@iit.bme.hu, hanak@emt.bme.hu').
-vsn('$LastChangedDate: 2020-09-14 19:57:47 +0200 (h, 14 szept 2020) $$').

% ----------------------- Bemelegítés ----------------------------------

-type fa() :: 1 | {fa(), '+', fa()}.
-spec fak(N :: integer()) -> F :: [fa()].
% Az összes N levelű fa listája F.
fak(1) ->
    [1];
fak(N) ->
    [ {BalFa, '+', JobbFa}
      || I <- lists:seq(1, N - 1),
         BalFa <- fak(I),
         JobbFa <- fak(N - I)
    ].

% ----------Adott értékű kifejezések felsorolása -----------------------

-type muvelet() :: '+' | '-' | '*' | '/'.
-type kif() :: integer() | {kif(), muvelet(), kif()}.
-spec kifek(Ls::[integer()]) -> Ks::[kif()].
% Az Ls-beli számokból az alapműveletekkel képzett kifejezések listája Ks.
% A levelek in-order sorrendje az Ls-beli sorrend.
% Megj.: Erlangban explicit felsoroljuk, nincs automatikus visszalépés.
kifek([H]) ->
    [H];
kifek(Ls) ->
    [ {B, M, J}
      || {BalLista, JobbLista} <- kettevagasok(Ls),
         B <- kifek(BalLista),
         J <- kifek(JobbLista),
         M <- ['+', '-', '*', '/']
    ].

-type int() :: integer().
-spec kettevagasok(Ls::[int()]) -> PLs::[{BLs::[int()], JLs::[int()]}].
% Az összes olyan nemüres BLs és JLs listákból álló párok listája PLs,
% amelyek páronként összefűzve az Ls listát adják.
kettevagasok(Ls) ->
    [ {BalLista, JobbLista}
      || I <- lists:seq(1, length(Ls) - 1),
         {BalLista, JobbLista} <- [lists:split(I, Ls)]
    ].

-spec ertek(K :: kif()) -> E :: integer().
% A K kifejezés értéke E.
% Megj.: Erlangban explicit el kell végezni a kiértékelést.
ertek({BalFa, Muvelet, JobbFa}) ->
    erlang:Muvelet(ertek(BalFa), ertek(JobbFa));
ertek(I) ->
    I.
% Megj.: a klózok sorrendje itt számít!

-spec permutaciok(Ls :: [any()]) -> Ps :: [[any()]].
% Az Ls lista elemeinek összes permutációját tartalmazó lista Ps.
permutaciok([]) ->
    [[]];
permutaciok(L) ->
    [[H|T] || H <- L, T <- permutaciok(L -- [H])].

-spec megoldasok() -> Ks :: [kif()].
% Az 1,3,4,6 számokkal a 24 eredményt adó kifejezések listája Ks.
megoldasok() -> megoldasok([1,3,4,6], 24).

-spec megoldasok(Ls :: [integer()], E :: integer()) -> Ks :: [kif()].
% Az Ls számokkal az E eredményt adó kifejezések listája Ks.
megoldasok(Szamok, Eredmeny) ->
    [Kif || L <- permutaciok(Szamok),
            Kif <- kifek(L),
            (catch ertek(Kif)) == Eredmeny].
%	    ertek(Kif) == Eredmeny].
% Megjegyzés
% catch: ha nincs hiba, az ertek(Kif) értékét adja eredményül; hiba esetén, pl.
% 0-val osztáskor „elkapja” a hibát, és egy ennest ad eredményül..

% ----------------------- Példák ---------------------------------------

pelda(1) ->
    [X || X <- [1,2,3]];
pelda(2) ->
    [X || X <- [1,2,3], X*X > 5];
pelda(3) ->
    lists:seq(1,3);
pelda(4) ->
    [{X,Y} || X <- [1,2,3,4], Y <- lists:seq(1,X)];
pelda(5) ->
    fak(4);
pelda(6) ->
    length(fak(4));
pelda(7) ->
    kifek([1,3,4]);
pelda(8) ->
    ertek({{1, '-', {3, '*', 4} }, '+', 6});
pelda(9) ->
    permutaciok([1,3,4]);
pelda(10) ->
    L = [-1,-2,-3,-4],
    [{Elso,Masodik} || I <- [1,2,3], {Elso,Masodik} <- [lists:split(I, L)]].
