-module(orok).
% -compile(export_all).
-export([fac/1, absz/1, kategoria/1, teszt/0]).
-author('kapolnai@iit.bme.hu, hanak@emt.bme.hu').
-vsn('$LastChangedDate: 2019-10-29 21:47:55 +0100 (k, 29 okt 2019) $$').

-spec fac(N::integer()) -> F::integer().
% F = N! (F az N faktoriálisa).
fac(0) ->
    1;
fac(N) when is_integer(N), N>0 ->
    N * fac(N-1).

-spec absz(Num::number()) -> Abs::number().
%% Num abszolút értéke Abs.
absz(X) when is_number(X), X<0 ->
    -X;
absz(X) when is_number(X) ->
    X.

-spec kategoria(V::term()) -> K::atom().
%% A V term itt definiált kategóriája K.
kategoria(V) ->
    case V of
        X when is_atom(X) ->
            atom;
        X when is_number(X), X < 0 ->
            negativ_szam;
        X when is_integer(X) ;
               is_float(X), abs(X-round(X)) < 0.0001 ->
            kerek_szam;
        X when is_list(X), length(X) > 5 ->
            hosszu_lista;
        {X,Y,Z} when X*X+Y*Y =:= Z*Z, is_integer(Z) ;
                     Z*Z+Y*Y =:= X*X, is_integer(X) ;
                     X*X+Z*Z =:= Y*Y, is_integer(Y) ->
            pitagoraszi_szamharmas;
        {Nev, []} when is_atom(Nev) ->
            talan_hallgato;
        {Nev, [{Tipus,_}|_]} when is_atom(Nev), is_atom(Tipus) ->
            talan_hallgato;
        [Ny1|_] when Ny1 =:= cekla ; Ny1 =:= prolog ; Ny1 =:= erlang ->
            talan_programozasi_nyelvek_listaja;
        {tort, Sz, N} when abs(Sz div N) >= 0 ->
	    % Ha Sz vagy N nem egész, vagy ha N=:=0, hiba miatt meghiúsul
            racionalis;
        _ -> egyeb
    end.

teszt() ->
    [{K,orok:kategoria(K)} || K <- [true, haho, -5, 5.000001, "kokusz",
                                    {3,5,4}, {'D.D.',[]}, {tort,1,a},
                                    {tort,1,2},[cekla,prolog,erlang]]
    ]
    .
