-module(set).
-compile(export_all).
-author('hanak@iit.bme.hu, modified by: kapolnai@iit.bme.hu').
-vsn('$LastChangedDate: 2020-10-05 10:02:09 +0200 (h, 05 okt 2020) $$').

%% Halmaz megvalósítása rendezetlen listával.

-type set() :: list().

-spec empty() -> E::set().
%% E az üres halmaz.
%% Az absztrakcióhoz célszerű, ábrázolástól független interfész.
%% Sajnos, nem tudunk következetesen elszakadni az ábrázolástól.
empty() ->
    [].

-spec isMember(X::any(), Ys::set()) -> B::boolean().
%% B igaz, ha az X eleme az Ys halmaznak.
isMember(_, []) ->
    false;
isMember(X, [Y|Ys]) ->
    X=:=Y orelse isMember(X, Ys).

-spec isMember2(X::any(), Ys::set()) -> B::boolean().
%% B igaz, ha az X eleme az Ys halmaznak.
isMember2(_, []) ->
    false;
isMember2(X, [X|_]) ->
    true;
isMember2(X, [_Y|Ys]) ->
    isMember2(X, Ys).

-spec newMember(X::any(), Xs::set()) -> Rs::set().
%% Az Rs halmaz az Xs halmaz és az [X] halmaz uniója.
newMember(X, Xs) ->
    case isMember(X, Xs) of
        true -> Xs;
        false -> [X|Xs]
    end.

-spec listToSet(Xs::list()) -> Rs::set().
%% Rs az Xs lista elemeinek ismétlődésmentes halmaza.
listToSet([]) ->
    empty();
listToSet([X|Xs]) ->
    newMember(X, listToSet(Xs)).

-spec union(Xs::set(), Ys::set()) -> Zs::set().
%% Zs az Xs és Ys halmazok uniója.
union([], Ys) ->
    Ys;
union([X|Xs], Ys) ->
    newMember(X, union(Xs, Ys)).

-spec union2(Xs::set(), Ys::set()) -> Zs::set().
%% Zs az Xs és Ys halmazok uniója.
union2(Xs, Ys) ->
    lists:foldr(fun newMember/2, Ys, Xs).

-spec union3(Xs::set(), Ys::set()) -> Zs::set().
%% Zs az Xs és Ys halmazok uniója.
union3(Xs, Ys) ->
    lists:foldl(fun newMember/2, Ys, Xs).

-spec intersect(Xs::set(), Ys::set()) -> Zs::set().
%% Zs az Xs és Ys halmazok metszete.
intersect([], _) ->
    [];
intersect([X|Xs], Ys) ->
    Zs = intersect(Xs, Ys),
    case isMember(X, Ys) of
        true ->
            [X|Zs];
        false ->
            Zs
    end.

-spec intersect2(Xs::set(), Ys::set()) -> Zs::set().
%% Zs az Xs és Ys halmazok metszete.
intersect2([], _) ->
    [];
intersect2([X|Xs], Ys) ->
    case isMember(X, Ys) of
        true ->
            [X|intersect2(Xs, Ys)];
        false ->
            intersect2(Xs, Ys)
    end.

-spec intersect3(Xs::set(), Ys::set()) -> Zs::set().
%% Zs az Xs és Ys halmazok metszete.
intersect3(Xs, Ys) ->
    [ X || X <- Xs, isMember(X, Ys) ].

-spec intersect4(Xs::set(), Ys::set()) -> Zs::set().
%% Zs az Xs és Ys halmazok metszete.
intersect4(Xs, Ys) ->
    Xs -- (Xs -- Ys).

-spec isSubset(Xs::set(), Ys::set()) -> B::boolean().
%% B igaz, ha Xs részhalmaza Ys-nek.
isSubset([], _) ->
    true;
isSubset([X|Xs], Ys) ->
    isMember(X, Ys) andalso isSubset(Xs, Ys).

-spec isEqual(Xs::set(), Ys::set()) -> B::boolean().
%% B igaz, ha Xs és Ys elemei azonosak.
isEqual(Xs, Ys) ->
    isSubset(Xs, Ys) andalso isSubset(Ys, Xs).

-spec powerSet1(Xs::set()) -> Zss::list(set()).
%% Zss az Xs halmaz hatványhalmaza.
powerSet1([]) ->
    [[]];
powerSet1([X|Xs]) ->
    P = powerSet1(Xs),
    P ++ [ [X|Ys] || Ys <- P ].

-spec powerSet2(Xs::set()) -> Zss::list(set()).
%% Zss az Xs halmaz hatványhalmaza.
powerSet2(Xs) ->
    lists:foldl(fun(X, Xss) ->
                        Xss ++ [ [X|Ys] || Ys <- Xss ]
                end,
                [[]],
                Xs).

-spec insAll(X::any(), Yss::[set()], Zss::[set()]) -> Xss::[set()].
%% Xss olyan [X|Ys]-ek Zss elé fűzött listája, amelyekben az Ys-ek az Yss elemei.

insAll(_X,[],Zss) ->
    Zss;
insAll(X,[Ys|Yss],Zss) ->
    insAll(X,Yss,[[X|Ys]|Zss]).

%% insAll2(X,Yss) ->
%%     [[X|Ys] || Ys <- Yss].

-spec powerSet3(Xs::set()) -> Zss::list(set()).
%% Zss az Xs halmaz hatványhalmaza.
powerSet3([]) ->
    [[]];
powerSet3([X|Xs]) ->
    P = powerSet3(Xs),
    insAll(X,P,P).

%%% Példák

stopper(F, A) ->
    erlang:garbage_collect(),timer:sleep(100),
    _ = statistics(runtime),
    R=F(A),
    {length(R),element(2,statistics(runtime))}.

test() ->
    [
     {isMember,isMember(f,[1,5,g,v,6,7,d,c,y,q,w,",","=",alma,f,akna])},
     {isMember2,isMember2(a,[1,5,g,v,6,7,d,c,y,q,w,",","=",alma,akna])},
     {newMember,newMember(5,[0,2,4,6,8])},
     {newMember,newMember(5,[1,3,5,7,9])},
     {listToSet,listToSet([1,2,3,4,5,6,7,8,9,8,6,4])},
     {union,union([6,4,2],[9,7,3,5])},
     {union,union([7,6,4,3,2],[9,7,3,5,6])},
     %% {union2,union2([6,4,2],[9,7,3,5])},
     %% {union2,union2([7,6,4,3,2],[9,7,3,5,6])},
     %% {union3,union3([6,4,2],[9,7,3,5])},
     %% {union3,union3([7,6,4,3,2],[9,7,3,5,6])},
     {intersect,intersect([6,4,2],[9,7,3,5])},
     {intersect,intersect([7,6,4,3,2],[9,7,3,5,6])},
     %% {intersect2,intersect2([6,4,2],[9,7,3,5])},
     %% {intersect2,intersect2([7,6,4,3,2],[9,7,3,5,6])},
     %% {intersect3,intersect3([6,4,2],[9,7,3,5])},
     %% {intersect3,intersect3([7,6,4,3,2],[9,7,3,5,6])},
     {isSubset,isSubset([2,7,6,3],[9,7,3,5,6])},
     {isSubset,isSubset([7,6,3],[9,7,3,5,6])},
     {isSubset,isSubset([9,7,3,5,6],[7,6,3])},
     {isEqual,isEqual([7,6,3],[6,3,7])},
     {isEqual,isEqual([7,6,3],[7,6,3])},
     {isEqual,isEqual([7,6,3],[7,6,3,0])},
     {powerSet1,powerSet1([1,2,3])},
     {length,length(powerSet1(lists:seq(1,10)))},
     {powerSet2,powerSet2([1,2,3])},
     {length,length(powerSet2(lists:seq(1,10)))},
     {powerSet3,powerSet3([1,2,3])},
     {length,length(powerSet3(lists:seq(1,10)))},
     begin
         Canon = fun(PS) -> lists:usort([lists:usort(S) || S <- PS(lists:seq(1,10))]) end,
         Canon(fun set:powerSet1/1) =:= Canon(fun set:powerSet3/1)
     end
%     lists:usort([lists:usort(S) || S <- set:powerSet(lists:seq(1,10))]) =:= lists:usort([lists:usort(S) || S <- set:powerSet3(lists:seq(1,10))])
    ].
